/*Copyright (c) 2019 Alexis Nasr && Franck Dary

 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:i

 The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.*/
/// \file util.hpp
/// \brief Contains a variety of useful functions that doesn't operate on custom classes.
/// \author Franck Dary
/// @version 2.0
/// @date 2019-12-10

#ifndef UTIL__H
#define UTIL__H

#include <string>
#include <vector>
#include <array>
#include <unordered_map>
#include <fmt/core.h>
#include <experimental/source_location>
#include "utf8.hpp"

namespace util
{
typedef std::array<char, 4> utf8char;
typedef std::basic_string<utf8char> utf8string;

std::string_view getFilenameFromPath(std::string_view s);

std::vector<std::string_view> split(std::string_view s, char delimiter);
utf8string splitAsUtf8(std::string_view s);

void warning(std::string_view message, const std::experimental::source_location & location = std::experimental::source_location::current());
void error(std::string_view message, const std::experimental::source_location & location = std::experimental::source_location::current());
void error(const std::exception & e, const std::experimental::source_location & location = std::experimental::source_location::current());
void myThrow(std::string_view message, const std::experimental::source_location & location = std::experimental::source_location::current());

std::string int2HumanStr(int number);

template<typename T>
std::size_t memorySize(const T & val)
{
  myThrow("Type not yet supported");
  return sizeof val;
}

inline std::size_t memorySize(int val)
{
  return sizeof val;
}

template<typename T>
std::size_t memorySize(const std::basic_string<T> & val)
{
  return sizeof val + val.capacity() * sizeof (T);
}

template<typename T>
std::size_t memorySize(const std::vector<T> & vec)
{
  std::size_t result = sizeof vec + sizeof (T) * (vec.capacity()-vec.size());

  for (auto & elem : vec)
    result += memorySize(elem);

  return result;
}

};

template <>
struct fmt::formatter<std::experimental::source_location>
{
  constexpr auto parse(format_parse_context & ctx) { return ctx.begin(); }

  template <typename FormatContext>
  auto format(const std::experimental::source_location & d, FormatContext & ctx)
  {
    return format_to(ctx.out(), "{},l.{},'{}'", util::getFilenameFromPath(d.file_name()), d.line(), d.function_name());
  }
};

template <>
struct fmt::formatter<util::utf8char>
{
  constexpr auto parse(format_parse_context & ctx) { return ctx.begin(); }

  template <typename FormatContext>
  auto format(const util::utf8char & c, FormatContext & ctx)
  {
    char * asPtr = (char*)&c;
    return format_to(ctx.out(), "{}{}{}{}", asPtr[0] ? asPtr[0] : '\0', asPtr[1] ? asPtr[1] : '\0', asPtr[2] ? asPtr[2] : '\0', asPtr[3] ? asPtr[3] : '\0');
  }
};

std::string_view operator+(std::string_view a, std::string_view b);
void operator+=(std::string_view & a, std::string_view b);

#endif
